#define VERSION "GSModeSelector_v0.21"

/*
# GS Mode Selector - Force (set and keep) a GS Mode, then load & exec a PS2 ELF
#------------------------------------------------------------------------------
# Copyright 2009, Doctorxyz
# Licenced under Academic Free License version 2.0
# Review LICENSE file for further details.
*
* Thanks goes to (in alphabetical order)...
* bootsector
* dariuszg
* darthkamikaz3
* david_wgtn
* dlanor
* don roberto
* euzeb2
* ffgriever
* gilberto moreira
* goriath
* guxtabo
* jepjepjep
* kevstah2004
* maverickhl
* rikimaru
* rodpad
* stefanol69
* tna
* urbigbro
* vegeta
* zin0099
* ... and all those that contributed (and still contribute) for this project!
*
*/

//#define PS2ELF1	"mc0:/BOOT/ESR.ELF"			// Put here the 1st. path and ELF file name option you want to boot after applying the patch
#define PS2ELF2	"mc0:/BOOT/HDLOADER.ELF"	// Put here the 2nd. path and ELF file name option you want to boot after applying the patch

#include <stdio.h>
#include <kernel.h>
#include <sifrpc.h>
#include <libhdd.h>
#include <debug.h>
#include <malloc.h>
#include <fileXio_rpc.h>
#include <loadfile.h>
#include <fileio.h>
#include <string.h>
#include <libmc.h>
#include <hw.h>
#include <libpad.h>

#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include "modules.h"
#include "timer.h"
#include "pad.h"

/* Prototypes for External Functions */
void RunLoaderElf(char *filename, char *);

#define MAKE_J(func)		(u32)( (0x02 << 26) | (((u32)func) / 4) )	// Jump (MIPS instruction)
#define NOP					0x00000000									// No Operation (MIPS instruction)
#define GET_LOW(val)	(u32)( val & 0xFFFFFFFF )						// Return a Word value containg the less significant part from a Double Word value
#define GET_HIGH(val)	(u32)( val >> 32 )								// Return a Word value containg the most significant part from a Double Word value

// TSR - Data Section
// ------------------

//Str_Exp Macro below helps us stringify expanded arguments of other macros
#define Str_Exp(stringparts) #stringparts

//KDef Macro below defines typed symbols for KSeg space with a given address
#define KDef(type, name, address)\
	extern type name;\
	__asm__(Str_Exp(name = address\n\t)) //Do not comment the two lines above !

//KRel Macro below defines typed symbols for KSeg space with a given offset
//relative to KSeg. The pure symbol is absolute address, and is available
//both in asm and in C, but a "rel_" prefix gives the offset, for asm only.
#define KRel(type, name, offset)\
	extern type name;\
	__asm__(Str_Exp(rel_##name = offset\n\t));\
	__asm__(Str_Exp(name = offset+KSeg\n\t)) //No comment on three lines above !

	KDef(void, KSeg, 0x00080000);  //KSeg is the base of our KSeg area
	KRel(u128, MIPS_Regs, 0x0000); //Here we save exception entry registers
	//NB: The MIPS_Regs area must be 16-byte aligned, for 128 bit MIPS registers
	//But our coding also requires MIPS_Regs == KSeg, with the lower 16 bits zero

	KRel(u32, Target_Interlace, 0x0200);
	KRel(u32, Target_Mode,      0x0204);
	KRel(u32, Target_Field,     0x0208);

	KRel(u32, Adapt_Flags,  0x020C);	//Four Byte flags combined
	KRel(u8,  Adapt_Flag_0, 0x020C);	//Byte flag
	KRel(u8,  Adapt_Flag_1, 0x020D);	//Byte flag
	KRel(u8,  Adapt_Flag_2, 0x020E);	//Byte flag
	KRel(u8,  Adapt_Flag_3, 0x020F);	//Byte flag

	KRel(u64, Target_Display, 0x0210);  //Should be u64, but used as u128

	KRel(u32,  Source_Interlace, 0x0220);
	KRel(u32,  Source_Mode,    0x0224);
	KRel(u32,  Source_Field,   0x0228);
	KRel(u64, Source_Display, 0x0230);  //Should be u64, but used as u128
//NB: Source_Display above is unsafe! Something zeroes it...
//NB: For this reason we now use Safe_Orig_display instead
	KRel(void*, syscallTable, 0x0240);  //Pointer to function vector table
	KRel(void*, sceSetGsCrt, 0x0244);	  //Pointer to function

	KRel(u64, Target_SyncV,   0x0248);
	KRel(u64, Target_SMode2,	0x0250);

	KRel(u64, Safe_Orig_Display, 0x0260); //Should be u64, but used as u128
	KRel(u64, Calc_Display, 0x0270);      //Should be u64, but used as u128

	KDef(void, KSEG_func_base, 0x00080280);

	KDef(u64, GS_SMODE2,   0x12000020);
	KDef(u64, GS_SYNCV,    0x12000060);
	KDef(u64, GS_DISPLAY1, 0x12000080);
	KDef(u64, GS_DISPLAY2, 0x120000A0);

// TSR - Code Section
// ------------------

void *KSEG_next_func_p = &KSEG_func_base;
void *KSEG_ModdedSetGsCrt_p;
void *KSEG_DisplayHandler_p;

int size_ModdedSetGsCrt;
int size_DisplayHandler;

u32 config;

/* ----------------------------------------------------------------------------------------------------------------- */
//Here start the function segments intended for KSEG

//----- Start of ModdedSetGsCrt stuff -----
void beg_ModdedSetGsCrt();
void end_ModdedSetGsCrt();

__asm__( //Starts asm block for ModdedSetGsCrt
	".set noreorder\n\t"
"beg_ModdedSetGsCrt:\n\t"
	"lui $v0, (KSeg >> 16)\n\t"	// Upper Immediate 8 = 0x80000 = Base Address
	"addiu $sp, $sp, -0x0010\n\t"		// reserve 16 bytes stack space (1 reg)
	"sw $a0, rel_Source_Interlace($v0)\n\t"
	"sw $a1, rel_Source_Mode($v0)\n\t"
	"sw $a2, rel_Source_Field($v0)\n\t"
	"sync\n\t"						// Wait until preceding stores complete
	"lw $a3, rel_sceSetGsCrt($v0)\n\t"		// a3 -> original SetGsCrt function
	"lw $a0, rel_Target_Interlace($v0)\n\t"
	"lw $a1, rel_Target_Mode($v0)\n\t"
	"lw $a2, rel_Target_Field($v0)\n\t"
	"sync\n\t"						// Wait until the preceding loads are completed
	"sd $ra, 0($sp)\n\t"	// Push return address on stack
	"jalr $a3\n\t"				// Call original SetGsCrt
	"nop\n\t" 
	"ld $ra, 0($sp)\n\t"	// Pull return address from stack 
	"jr $ra\n\t"					// Return to caller
	"addiu $sp, $sp, 0x0010\n\t"	// Restore sp during return
"end_ModdedSetGsCrt:\n\t"
	".set reorder\n\t"                                    
);//ends asm block for ModdedSetGsCrt

void prep_ModdedSetGsCrt(){
	KSEG_ModdedSetGsCrt_p = KSEG_next_func_p;
	size_ModdedSetGsCrt = end_ModdedSetGsCrt-beg_ModdedSetGsCrt;
	KSEG_next_func_p += ((size_ModdedSetGsCrt+15)& -16);
}
//----- End of ModdedSetGsCrt stuff -----

//----- Start of DisplayHandler stuff -----
void beg_DisplayHandler();
void end_DisplayHandler();

__asm__( //Starts asm block for DisplayHandler
	".set noreorder\n\t"
	".set noat\n\t"

"beg_DisplayHandler:\n\t"
	"sq      $k0, -0x30($zero)\n\t"		// Store registers reserved for kernel
	"sq      $k1, -0x40($zero)\n\t"		// usage in interrupt/trap handling 
	//Save all MIPS registers except zero($0) k0($26) and k1($27)) into MIPSRegisters array
	"lui $k0, (MIPS_Regs >> 16)\n\t"
	"sq	$1, 0x10($k0)\n\t"
	"sq	$2, 0x20($k0)\n\t"
	"sq	$3, 0x30($k0)\n\t"
	"sq	$4, 0x40($k0)\n\t"
	"sq	$5, 0x50($k0)\n\t"
	"sq	$6, 0x60($k0)\n\t"
	"sq	$7, 0x70($k0)\n\t"
	"sq	$8, 0x80($k0)\n\t"
	"sq	$9, 0x90($k0)\n\t"
	"sq	$10, 0xA0($k0)\n\t"
	"sq	$11, 0xB0($k0)\n\t"
	"sq	$12, 0xC0($k0)\n\t"
	"sq	$13, 0xD0($k0)\n\t"
	"sq	$14, 0xE0($k0)\n\t"
	"sq	$15, 0xF0($k0)\n\t"
	"sq	$16, 0x100($k0)\n\t"
	"sq	$17, 0x110($k0)\n\t"
	"sq	$18, 0x120($k0)\n\t"
	"sq	$19, 0x130($k0)\n\t"
	"sq	$20, 0x140($k0)\n\t"
	"sq	$21, 0x150($k0)\n\t"
	"sq	$22, 0x160($k0)\n\t"
	"sq	$23, 0x170($k0)\n\t"
	"sq	$24, 0x180($k0)\n\t"
	"sq	$25, 0x190($k0)\n\t"
	"sq	$28, 0x1C0($k0)\n\t"
	"sq	$29, 0x1D0($k0)\n\t"
	"sq	$30, 0x1E0($k0)\n\t"
	"sq	$31, 0x1F0($k0)\n\t"

	"mfc0	$k0, $30\n\t"	//decode MIPS instruction from ErrorPC (COP0 reg: $30),
											// get rt (target register) and write address
	"lw	$t0, 0($k0)\n\t"			// $t0 = MIPS instruction (trapped by Data Address Write Breakpoint)
	"srl	$t1, $t0, 16\n\t"			
	"andi	$t1, $t1, 0x1f\n\t"		// $t1 = unscaled rt reg index
	"srl	$t3, $t0, 21\n\t"
	"andi	$t3, $t3, 0x1f\n\t"		// $t3 = unscaled base reg index

	"sll	$k0, $t3, 4\n\t"			// $k0 = raw base_ix << 4 (scaled base_ix reg index)

	"lui $t6, (MIPS_Regs >> 16)\n\t"	// t6 -> MIPS_Regs
	"addu	$t6, $t6, $k0\n\t"		// $t6 = &MIPS_Regs[base_ix]; (if type = u128)
	"lw	$t4, 0($t6)\n\t"				// $t4 = base register value
	"li	$k0, 0xffff\n\t"				// $k0 = mask for 16 bits
	"and	$k1, $t0, $k0\n\t"		// $k1 = offset field of instruction
	"addu	$t4, $t4, $k1\n\t"		// $t4 = address which triggered breakpoint

	"sll	$k0, $t1, 4\n\t"			// $k0 = raw rt_ix << 4 (scaled rt_ix reg index)

	"lui $t5, (MIPS_Regs >> 16)\n\t"	// t5 -> MIPS_Regs
	"addu	$t5, $t5, $k0\n\t"		// $t5= &MIPS_Regs[rt_ix];
	"lq	$t2, 0($t5)\n\t"				// $t2 = value in rt

	"lui $t8, (KSeg >> 16)\n\t"		// t8 -> KSeg
	"sq $t2, rel_Safe_Orig_Display($t8)\n\t"	//Used to be Safe_Orig_Display

	"lq $t7, rel_Target_Display($t8)\n\t"			// Load target parameter value from Target_Display


// ----------------------------------------------------------------------------------------------------------------------------------------------
// Source -> Original values for DX, DY, MAGH, MAGV, DW and DH (DISPLAYx GS Registers) fields
// Target -> Modded(forced) values for DX', DY', MAGH', MAGV', DW' and DH' (DISPLAYx GS Registers) fields
// DISPLAYx is the source value for DISPLAY1/DISPLAY2 (it depends on the context).
// and DISPLAYx' is the target value to be forced.

//Patch to adapt request to enforced mode in t7 MUST also preserve t4, t8
	".set at\n\t"               		// We must enable $at for DIV use
	"li	$v0,0\n\t"						// preclear v0 as result DISPLAYx accumulator
	"lq $t9,rel_Safe_Orig_Display($t8)\n\t"	// t9=request DISPLAYx value
	"lq $t7,rel_Target_Display($t8)\n\t"    // t7=forcing DISPLAYx template

	"lb $t0,rel_Adapt_Flag_0($t8)\n\t"
	"bnel	$t0,$zero,91f\n\t"						//if(Adapt_Flag_0)
	"or $t9,$zero,$t7\n\t"      				//	simulate request same as forced mode
"91:\n\t"

	"lb $t0,rel_Adapt_Flag_1($t8)\n\t"
	"bnel	$t0,$zero,92f\n\t"  				//if(Adapt_Flag_1) Skip_adaption_work
	"or $v0,$zero,$t9\n\t"      			//	Instead copying the request as-is

#include	"Adapt_X.c"
#include	"Adapt_Y.c"

"92:\n\t"
	"sq $v0,rel_Calc_Display($t8)\n\t"  //Store new DISPLAYx value (for feedback)

	"or $t7,$zero,$v0\n\t"				//Move accumulated result to t7
//End of Patch to adapt request with the resulting request in t7

//----- Start of: Additional register patching added from v0.21 -----
/*SMODE2
.----.---.---------.-----------------------------------.
|Name|Pos|Format   |Contents                           |
+----+---+---------+-----------------------------------|
|INT | 0 |int 0:1:0|Interlace Mode Setting             |
|    |   |         |0 Non-Interlace Mode               |
|    |   |         |1 Interlace Mode                   |
|FFMD| 1 |int 0:1:0|Setting in Interlace Mode          |
|    |   |         |0 FIELD Mode(Read every other line)|
|    |   |         |1 FRAME Mode(Read every line)      |
|DPMS|3:2|int 0:2:0|VESA DPMS Mode Setting             |
|    |   |         |00 On          10 Suspend          |
|    |   |         |01 Stand-by    11 Off              |
^----^---^---------^-----------------------------------.*/

	"ld $t0, rel_Target_SMode2($t8)\n\t"
	"beqz $t0, 93f\n\t"				// Don't update SMODE2 register with zero value
	"nop\n\t"
 	"sd $t0, GS_SMODE2\n\t"			// Store modified value to GS SMODE2 register
"93:\n\t"

	"lb $t2,rel_Adapt_Flag_2($t8)\n\t"
	"bnel	$t2,$zero,94f\n\t"							//if(Adapt_Flag_2)
	"lq $t7, rel_Target_Display($t8)\n\t"	//	use forced mode without adaption
"94:\n\t"

	"sd	$t7, GS_DISPLAY2\n\t"			// Store modified value to GS DISPLAY2 register
	"sd	$t7, GS_DISPLAY1\n\t"			// Store modified value to GS DISPLAY1 register
//	"sd	$t7, 0($t4)\n\t"				// Store modified value to GS DISPLAYx register

/*SYNCV
.----.-----.----------.
|Name|Pos. |Format    |
|----+-----+----------+
|VFP | 9:0 |int 0:10:0|
|VFPE|19:10|int 0:10:0|
|VBP |31:20|int 0:12:0|
|VBPE|41:32|int 0:12:0|
|VDP |52:42|int 0:11:0|
|VS  |??:53|int 0:??:0|
'----^-----^----------^*/

	"ld $t1, rel_Target_SyncV($t8)\n\t"
	"beqz $t1, 95f\n\t"					// Don't update SYNCV register with zero value
	"nop\n\t"
 	"sd $t1, GS_SYNCV\n\t"			// Store modified value to SYNCV register
"95:\n\t"

//----- End of: Additional register patching added from v0.21 -----
//The above also replaces the commented storage instruction immediately below
//	"sq	$t7, 0($t4)\n\t"			// Store modified value to GS DISPLAYx register
															// $t7 = new value for rt
	".set noat\n\t"
	"lui $k0, 0x0008\n\t"				// Restore all MIPS registers' Context (but zero($0) k0($26) and k1($27)) from MIPSRegisters array
	"lq	$1, 0x10($k0)\n\t"
	"lq	$2, 0x20($k0)\n\t"
	"lq	$3, 0x30($k0)\n\t"
	"lq	$4, 0x40($k0)\n\t"
	"lq	$5, 0x50($k0)\n\t"
	"lq	$6, 0x60($k0)\n\t"
	"lq	$7, 0x70($k0)\n\t"
	"lq	$8, 0x80($k0)\n\t"
	"lq	$9, 0x90($k0)\n\t"
	"lq	$10, 0xA0($k0)\n\t"
	"lq	$11, 0xB0($k0)\n\t"
	"lq	$12, 0xC0($k0)\n\t"
	"lq	$13, 0xD0($k0)\n\t"
	"lq	$14, 0xE0($k0)\n\t"
	"lq	$15, 0xF0($k0)\n\t"
	"lq	$16, 0x100($k0)\n\t"
	"lq	$17, 0x110($k0)\n\t"
	"lq	$18, 0x120($k0)\n\t"
	"lq	$19, 0x130($k0)\n\t"
	"lq	$20, 0x140($k0)\n\t"
	"lq	$21, 0x150($k0)\n\t"
	"lq	$22, 0x160($k0)\n\t"
	"lq	$23, 0x170($k0)\n\t"
	"lq	$24, 0x180($k0)\n\t"
	"lq	$25, 0x190($k0)\n\t"
	"lq	$28, 0x1C0($k0)\n\t"
	"lq	$29, 0x1D0($k0)\n\t"
	"lq	$30, 0x1E0($k0)\n\t"
	"lq	$31, 0x1F0($k0)\n\t"

	"mfc0	$k0, $30\n\t"				// ErrorEPC
	"addiu	$k0, $k0, 4\n\t"			// Increment ErrorEPC
	"mtc0	$k0, $30\n\t"
	"sync\n\t"

	"mfc0	$k0, $12\n\t"				// Set user mode, interrupts on
	"ori	$k0, $k0, 0x11\n\t"
	"mtc0	$k0, $12\n\t"
	"sync\n\t"

	"lq      $k0, -0x30($zero)\n\t"		// Restore registers reserved for OS Kernel ($k0 and $k1)
	"lq      $k1, -0x40($zero)\n\t"
	"sync\n\t"

	"eret\n\t"							// 	Return from exception
"end_DisplayHandler:\n\t"

	".set at\n\t"                                    
	".set reorder\n\t"                                    
);//Ends asm block for DisplayHandler

void prep_DisplayHandler(){
	KSEG_DisplayHandler_p = KSEG_next_func_p;
	size_DisplayHandler = end_DisplayHandler-beg_DisplayHandler;
	KSEG_next_func_p += ((size_DisplayHandler+15)& -16);
}
//----- End of DisplayHandler stuff -----

//Here end the function segments intended for KSEG

//----------------------------------------------------------------------------
u64 make_display_magic_number(dh, dw, magv, magh, dy, dx)
{
	u64 rv;

	rv = ((u64)dx      & ((1<<12)-1))
	    |(((u64)dy     & ((1<<11)-1))<<12)
	    |(((u64)magh   & ((1<<4)-1)) <<23)
	    |(((u64)magv   & ((1<<2)-1)) <<27)
	    |((((u64)dw-1) & ((1<<12)-1))<<32)
	    |((((u64)dh-1) & ((1<<11)-1))<<44);

//RA NB: dw and dh are here 1 unit higher than the real GS register values
//RA NB: so for a 480p resolution we use 480, though the register gets 479
//RA NB: which is thus what we here embed in the bits of the return value

	return rv;
}//ends make_display_magic_number
//----------------------------------------------------------------------------
unsigned int get_sr() {
    register unsigned int rv;
	__asm__ __volatile__ (
	".set noreorder\n\t"
    "mfc0 %0, $12\n\t"
    "sync\n\t"
    "nop\n\t"
	".set reorder\n\t"
    : "=r" (rv) : );
    return rv;
}

//----------------------------------------------------------------------------
void set_sr(unsigned int v) {
	__asm__ __volatile__ (
	".set noreorder\n\t"
    "mtc0 %0, $12\n\t"
    "sync\n\t"
    "nop\n\t"
	".set reorder\n\t"
    : : "r" (v) );
}

//----------------------------------------------------------------------------
void set_kmode() {
	set_sr(get_sr() & (~0x19));
}

//----------------------------------------------------------------------------
void set_umode() {
	set_sr((get_sr() & (~0x19)) | 0x11);
}

//----------------------------------------------------------------------------
int disableClearUserMem()
{
	int *scanMem = (int *)0x80000000;
	int retval = -1;
	int i;
	set_kmode();
	for(i = 0; i < 0x3FFFF; i++)
	{
	 if( ((scanMem[i] & 0xffff0000) == 0x3C040000) &&	 	// lui a0, 0000
		 ((scanMem[i + 1] & 0x7C000000) == 0xC000000) &&	// jal XXXX
		 (scanMem[i + 2] == 0x34842000) )				 	// ori a0, a0, 0x2000
		 {
			 scanMem[i + 1] = NOP;
			 retval = 0;
			 break;
		 }
	}
	set_umode();
	return retval;
}

//----------------------------------------------------------------------------
void delay(int count)
{
	int i;
	int ret;
	for (i = 0; i < count; i++) {
		ret = 0x01000000;
			while(ret--) asm("nop\nnop\nnop\nnop");
	}
}

//----------------------------------------------------------------------------
void poweroffHandler(int i)
{
   hddPowerOff();
}

//----------------------------------------------------------------------------
/* Read IOP modules */
void LoadModules(void) {
	/* Load memory card driver (IOP modules) - WARNING: Do not change this calling order! */
	SifLoadModule("rom0:SIO2MAN", 0, NULL);	// Basic serial communications module
	SifLoadModule("rom0:MCMAN", 0, NULL);	// MC (Memory card) control module #1
	SifLoadModule("rom0:MCSERV", 0, NULL);	// MC module #2
	mcInit(MC_TYPE_MC);						// Initialize MC driver
	SifLoadModule("rom0:PADMAN", 0, NULL);	// Controller control module
}

//----------------------------------------------------------------------------
/* This allow KSEG (Kernel EE RAM Area) Memory Copy */
u32 KSEG_memcpy(void *addr, void *buf, u32 size)
{
   DI();						// Disable interrupts
   set_kmode();					// Enter kernel mode
   memcpy(buf, addr, size);		// Memory Copy
   set_umode();					// Leave kernel mode
   EI();						// Enable interrupts
   return size;
}

//----------------------------------------------------------------------------
/*  Output Memory Dump to Screen - I recognize I have no skill, no time, and no conditions to learn / use naplink
    So I using this piece of trash code to debug the results on screen
    (Always when needed, I can take photos of it using my mobile and change code later) */
void MemoryDump(u32 Start, u32 Quantity)
{
        u32 i;
        u8 Data;

		scr_printf("[%08x] ", Start);		// Display Start Address
        
        for (i = 0; i < Quantity; i++)
        {
        	set_kmode();
        	Data = _lb(Start + i);
        	set_umode();
			scr_printf("%02x ", Data);		// Display Address Data
        }
		scr_printf("\n");					// Line break
}

//----------------------------------------------------------------------------
/* Get Syscall Vector Table Address - The value is different among PS2 versions, this is the reason why we must search for that! */
u32 GetSyscallVectorTableAddress(void)
{
   // Here, we set magical numbers (instead of ordinary address) to system call vector table positions 0xFE and 0xFF 
   // (both are related to syscalls code numbers that are unused). Doing this, i.e., puting non-usual opcodes sequence, 
   // that probably does not exist on original code, let us use this magic numbers (fake addresses) to found
   // Syscall Vector Table Start Address. It is used "BIG BOOBS CODE" here (0x1337C0DE+0xB16B00B5) but it could be
   // anything else that fits on this concept!
   const s32 syscall_num[2] = { 0xFE, 0xFF };
   const u32 magic[2] = { 0x1337C0DE, 0xB16B00B5 };	
   u32 addr = -1;
   u32 i;
   SetSyscall(syscall_num[0], (void*)magic[0]);
   SetSyscall(syscall_num[1], (void*)magic[1]);
   set_kmode();
   // Scan start address
   for (i = 0x80000000; i < 0x80080000; i += 4) {
      if (!memcmp((u32*)i, magic, sizeof(magic))) {
         addr = i - syscall_num[0] * 4;
         break;
      }
   }
   set_umode();
   SetSyscall(syscall_num[0], (void*)0);
   SetSyscall(syscall_num[1], (void*)0);
   return addr;
} 

//----------------------------------------------------------------------------
/* Set Syscall */
void setSyscall(int number, void (*functionPtr)(void)) {
	__asm__ __volatile__ (
	".set noreorder\n\t"
	".set noat\n\t"
	"li $3, 0x74\n\t"
    "add $4, $0, %0    \n"   // Specify the argument #1
    "add $5, $0, %1    \n"   // Specify the argument #2
   	"syscall\n\t"
	"jr $31\n\t"
	"nop\n\t"
	".set at\n\t"
	".set reorder\n\t"
    :
    : "r"( number ), "r"( functionPtr )
    );
}

//----------------------------------------------------------------------------
void setNormalAdaption(void) //Should be called before exiting main program
{
	set_kmode();
	Adapt_Flags &= 0xFFFF0000; //Clear 2 byte flags of 4
  set_umode();

}

//----------------------------------------------------------------------------
/* Install GS (Graphics Sinthesizer) Mode Selector */
void installGSModeSelector(u32 interlace, u32 mode, u32 field, u64 display, u64 syncv, u64 smode2)
{
	u32 SyscallVectorTableAddress;
	u32 SetGsCrtVectorAddress;
	u32 SetGsCrt_Addr;

	SyscallVectorTableAddress = GetSyscallVectorTableAddress();
	SetGsCrtVectorAddress = SyscallVectorTableAddress + 2 * 4;
	
	scr_printf("Syscall Vector Table found at 0x%08x\n", SyscallVectorTableAddress);
	MemoryDump(SyscallVectorTableAddress, 0x0C);
	   
	/* Install sceGsSetCrt patcher */
	/* Syscall hooked: Syscall #2 (SetGsCrt) */
	set_kmode();
	Adapt_Flags = 0x00000001;						//Init 4 byte flags
	Target_Interlace = interlace;
	Target_Mode = mode;
	Target_Field = field;
  set_umode();
  	
	scr_printf("Stored parameters to be forced by ModdedSetGstCrt (Interlace, Mode, Field) at\n");
	MemoryDump((u32) &Target_Interlace, 0x0C);
	
	set_kmode();
	Target_Display = display;
	((u64 *) &Target_Display)[1] = 0;
	set_umode();

	scr_printf("Stored parameters to be forced by DisplayHandler (Display) at\n");
	MemoryDump((u32) &Target_Display, 0x08);
	
	set_kmode();
	Target_SyncV = syncv;
	set_umode();

	scr_printf("                                                 (SyncV) at\n");
	MemoryDump((u32) &Target_SyncV, 0x08);

	set_kmode();
	Target_SMode2 = smode2;
	set_umode();

	scr_printf("                                                 (SyncV) at\n");
	MemoryDump((u32) &Target_SMode2, 0x08);

	set_kmode();
	syscallTable = (void *) SyscallVectorTableAddress;
    SetGsCrt_Addr = _lw(SetGsCrtVectorAddress);
		if(SetGsCrt_Addr == (int)KSEG_ModdedSetGsCrt_p){ //if already installed
			//Here we need to abort install, as GSM is already active.
			//The new instance of the GUI can use the old KSEG code directly
			set_umode();
			return; //abort recursive install
		}
		sceSetGsCrt = (void *) SetGsCrt_Addr;
	_sw((int)KSEG_ModdedSetGsCrt_p, SetGsCrtVectorAddress);	//SyscallVectorTableAddress[2] = (int)KSEG_ModdedSetGsCrt_p
	set_umode();

	scr_printf("[BEFORE]SyscallVectorTableAddress[2] =       SetGsCrt = 0x%08x\n", SetGsCrt_Addr);
	scr_printf("[AFTER ]SyscallVectorTableAddress[2] = ModdedSetGstCrt = 0x%08x\n\n", (int)KSEG_ModdedSetGsCrt_p);
	MemoryDump(SyscallVectorTableAddress, 0x0C);

	scr_printf("[BEFORE] Core Debug Exception Handler (V_DEBUG handler)\n");
	MemoryDump(0x80000100, 0x08);
	/* Install Display Handler in place of the Core Debug Exception Handler (V_DEBUG handler) */
	/* Exception Vector Address for Debug Level 2 Exception when Stadus.DEV bit is 0 (normal): 0x80000100 */
	/* 'Level 2' is a generalization of Error Level (from previous MIPS processors) */
	
	set_kmode();
	/* Do you know that '*((unsigned int*)Adddr) = Data' has the same effect that '_sw(Data), Address)'? ;-b */
	//*((unsigned int*)0x80000100) = MAKE_J((int)KSEG_DisplayHandler_p);
	//*((unsigned int*)0x80000104) = 0;
	_sw(MAKE_J((int)KSEG_DisplayHandler_p), 0x80000100);
	_sw(0, 0x80000104);
	//flush_cache();
	set_umode();
	scr_printf("[AFTER ] DisplayHandler -> J 0x%08x\n", (int)KSEG_DisplayHandler_p);
	MemoryDump(0x80000100, 0x08);
    
	/* Set Data Address Write Breakpoint */
	/* Trap writes to 0x12000080 (DISPLAY1) and 0x120000A0 (DISPLAY2)
	   in order to force their values (DX, DY, DW, DH, MAGH, MAGV) */
	__asm__ __volatile__ (
	".set noreorder\n\t"
	".set noat\n\t"
	
	"li $a0, 0x12000080\n\t"		// Address
	"li $a1, 0xFFFFFFDF\n\t"		// Mask
									// Will be trapped writes to 0x12000080 and 0x120000A0
	"di\n\t"						// Disable Interupts
	"sync\n\t"						// Wait until the preceding loads are completed

	"ori $k0, $zero, 0x8000\n\t"
	"mtc0 $k0, $24\n\t"				// All breakpoints off (BED = 1)
	"sync.p\n\t"					// Wait until the preceding instruction is completed with the exception of multiply, divide, multicycle COP1 or COP2 operations or a pending load.

//	I do not know why...
//	"mtc0 $a0, $24\n\t"
//	"mtc0 $a1, $24\n\t"

//	genenerates		00 C0 84 40
//					00 C0 85 40

//	instead of		04 C0 84 40
//					05 C0 85 40

//	So, I am using
//	"mtdab	$a0\n\t"
//	"mtdabm	$a1\n\t"

//	to generates	04 C0 84 40
//					05 C0 85 40

// and now, everything is ok!

	"mtdab	$a0\n\t"
	"mtdabm	$a1\n\t"


	"sync.p\n\t"					// Wait until the preceding instruction is completed with the exception of multiply, divide, multicycle COP1 or COP2 operations or a pending load.

	"mfc0 $k1, $24\n\t"
	"sync.p\n\t"					// Wait until the preceding instruction is completed with the exception of multiply, divide, multicycle COP1 or COP2 operations or a pending load.

	"lui $k0, 0x2020\n\t"			// Data write breakpoint on (DWE, DUE = 1)
	"or $k1, $k1, $k0\n\t"
	"xori $k1, $k1, 0x8000\n\t"		// DEBUG exception trigger on (BED = 0)
	"mtc0 $k1, $24\n\t"
	"sync.p\n\t"					// Wait until the preceding instruction is completed with the exception of multiply, divide, multicycle COP1 or COP2 operations or a pending load.

	"ei\n\t"						// Enable Interupts
//	"jr $ra\n\t"
	"nop\n\t"
	"nop\n\t"
	"nop\n\t"
	"nop\n\t"
	
	".set at\n\t"
	".set reorder\n\t"
	);
	scr_printf("Now, writes to DISPLAY1(0x12000080) and DISPLAY2(0x120000A0) are trapped!\n\n");

}

//----------------------------------------------------------------------------
/* Update parameters to be forced by ModdedSetGsCrt and DisplayHandler TSR functions */
void ModdedSetGsCrtDisplayHandlerParams(u32 interlace, u32 mode, u32 field, u64 display, u64 syncv, u64 smode2)
{
	set_kmode();
	Target_Interlace = interlace;
	Target_Mode = mode;
	Target_Field = field;
	Target_Display = display;
	((u64 *) &Target_Display)[1] = 0;
	Target_SyncV = syncv;
	Target_SMode2 = smode2;
  set_umode();
}
//----------------------------------------------------------------------------
void	CleanUp(void)
{
	setNormalAdaption();
	TimerEnd();
	padPortClose(1,0);
	padPortClose(0,0);
	padEnd();
}
//----------------------------------------------------------------------------
/*
+------------------------------+
|                              |
|   M A I N   F U N C T I O N  |
|                              |
+------------------------------+
*/
int main()
{   

	char fullpath[] = PS2ELF2;
	char party[1] = {0};
   
	u32 size;
	
	u32 interlace;
	u32 mode;
	u32 field;
	u64 display;
	u64 syncv;
	u64 smode2;
	
	char *on_off_s[] = {"ON", "OFF"};

	u64 presets, dispreq, dispcalc;
	
	u32 gs_dx, gs_dy, gs_magh, gs_magv, gs_dw, gs_dh;
//RA NB: gs_dw and gs_dh are 1 unit higher than the real GS register values
//RA NB: so for a 480p resolution we use 480, though the register gets 479

	int updateflag, adapt2;

	int retval, joy_step;

	/* PS2 Configuration settings
			note: the 'set' methods are only valid till the ps2 gets
			turned off or reset! */
	   
	/* Initialise */
//	SifResetIop();
   
	init_scr();
	
	scr_printf("%s - by Doctorxyz\n",VERSION);
	scr_printf("--------------------------------------------------------------------\n\n");
		
	scr_printf("Loading access modules...\n");
   	LoadModules();

 	scr_printf("Initialising the joypad subsystem...\n");
	TimerInit();
	Timer();
	setupPad();
	TimerInit();
	Timer();

	prep_ModdedSetGsCrt(); //Reserve KSEG space for ModdedSetGsCrt block
	scr_printf("Copying ModdedSetGstCrt resident function to KSEG0...");
	size = KSEG_memcpy(beg_ModdedSetGsCrt, KSEG_ModdedSetGsCrt_p, size_ModdedSetGsCrt);
	scr_printf("0x%08x bytes\n", size);

	prep_DisplayHandler(); //Reserve KSEG space for DisplayHandler block
	scr_printf("Copying DisplayHandler resident function to KSEG0...");
	size = KSEG_memcpy(beg_DisplayHandler, KSEG_DisplayHandler_p, size_DisplayHandler);
	scr_printf("0x%08x bytes\n", size);
	   
 	scr_printf("Disabling InitializeUserMemory function call in EE Kernel...\n");
	disableClearUserMem();
	
	/* The following interlace and field are the usual settings for non-interlaced modes! */
	interlace = 0x00;							// Non-interlaced -> Obtained from init_scr function from ps2dev\ps2lib\ee\libkernel\graph.c
	field = 0x01;								// Frame mode (Read Every Line) -> Obtained from init_scr function from ps2dev\ps2lib\ee\libkernel\graph.c

	mode = ((*((char*)0x1FC7FF52))=='E')+2;		// Auto -> Obtained from init_scr function from from ps2dev\ps2lib\ee\libkernel\graph.c
//RA NB: That method of testing for PAL/NTSC is valid only on old PS2 consoles
//RA NB: On the new PSTwo consoles it is necessary to read rom0:ROMVER instead



//RA NB: I'm adding PAL/NTSC dependent 'display' calculation here
	if(mode == 0x02){ //if initial mode is NTSC
		interlace = 0x01;			// Interlaced
		field = 0x00;					// Field mode (Read Every Other Line)
		gs_dx=700; gs_dy=46; gs_magh=3; gs_magv=0;
		gs_dw = 2560; gs_dh = 448;
	}else{ //else initial mode is PAL
		interlace = 0x01;			// Interlaced
		field = 0x00;					// Field mode (Read Every Other Line)
		gs_dx=720; gs_dy=68; gs_magh=3; gs_magv=0;
		gs_dw = 2560; gs_dh = 512;
	}

	display = make_display_magic_number(gs_dh, gs_dw, gs_magv, gs_magh, gs_dy, gs_dx);

	syncv = 0;		//Disable SYNCV register patching by default
	smode2 = 1;		//set SMODE2 for interlace by default

	/* Install patcher function for sceGsSetCrt and DISPLAY 1/2 GS Registers */
	scr_printf("Installing GS Mode Selector...\n");
	installGSModeSelector(interlace, mode, field, display, syncv, smode2);

	/* Just for debbuging purposes */
	delay(5);
	scr_printf("Take a photo and press any button to continue!\n");
	while(!(waitAnyPadReady(), readpad(), new_pad)); //await a new button

	updateflag = 1;
	while (!(updateflag == 0)) {

		presets = make_display_magic_number(gs_dh, gs_dw, gs_magv, gs_magh, gs_dy, gs_dx);
		if(updateflag == 1){
			display = presets;
			ModdedSetGsCrtDisplayHandlerParams(interlace, mode, field, display, syncv, smode2);
			scr_printf("This apparently useless printf is needed before the following init_scr. Do not ask me why! ;-)\n");
			init_scr();
		}//ends if
			
		set_kmode();
		adapt2 = Adapt_Flag_2;
		dispreq = Safe_Orig_Display;
		dispcalc = Calc_Display;
		set_umode();

		scr_clear();
					
		scr_printf("%s - by Doctorxyz\n",VERSION);
		scr_printf("--------------------------------------------------------------------\n\n");
		scr_printf("----- Current settings -----\n");
		scr_printf("Interlace %01d, Mode 0x%02x, Field %01d, Display 0x%08lx\n", interlace, mode, field, display);
		scr_printf("(DX=%04d,DY=%04d,MAGH=%02d,MAGV=%01d,DW=%04d,DH=%04d)\n", \
		(u16)((display >> 00) & 0x0FFF), \
		(u16)((display >> 12) & 0x07FF), \
		(u16)((display >> 23) & 0x000F), \
		(u16)((display >> 27) & 0x0003), \
		(u16)((display >> 32) & 0x0FFF), \
		(u16)((display >> 44) & 0x07FF));
		scr_printf("----- Latest dispreq== 0x%08lx-----\n", dispreq);
		scr_printf("(DX=%04d,DY=%04d,MAGH=%02d,MAGV=%01d,DW=%04d,DH=%04d)\n", \
		(u16)((dispreq >> 00) & 0x0FFF), \
		(u16)((dispreq >> 12) & 0x07FF), \
		(u16)((dispreq >> 23) & 0x000F), \
		(u16)((dispreq >> 27) & 0x0003), \
		(u16)((dispreq >> 32) & 0x0FFF), \
		(u16)((dispreq >> 44) & 0x07FF));
		scr_printf("----- Latest dispcalc== 0x%08lx-----\n", dispcalc);
		scr_printf("(DX=%04d,DY=%04d,MAGH=%02d,MAGV=%01d,DW=%04d,DH=%04d)\n", \
		(u16)((dispcalc >> 00) & 0x0FFF), \
		(u16)((dispcalc >> 12) & 0x07FF), \
		(u16)((dispcalc >> 23) & 0x000F), \
		(u16)((dispcalc >> 27) & 0x0003), \
		(u16)((dispcalc >> 32) & 0x0FFF), \
		(u16)((dispcalc >> 44) & 0x07FF));
		scr_printf("----- Current presets== 0x%08lx-----\n", presets);
		scr_printf("(DX=%04d,DY=%04d,MAGH=%02d,MAGV=%01d,DW=%04d,DH=%04d)\n", \
		(u16)((presets >> 00) & 0x0FFF), \
		(u16)((presets >> 12) & 0x07FF), \
		(u16)((presets >> 23) & 0x000F), \
		(u16)((presets >> 27) & 0x0003), \
		(u16)((presets >> 32) & 0x0FFF), \
		(u16)((presets >> 44) & 0x07FF));
/*
		scr_printf("new_pad=0x%08X paddata=0x%08X joy_value=%+02d joy_flags=0x%04X\n\n", \
		new_pad, paddata, joy_value, joy_flags);
*/

		scr_printf("----------PreDef Settings----------\n");
		scr_printf("[O] + [<]/[>] NTSC/HDTV 480i/480p  [A] + [<]/[>]  PAL/HDTV 576i/576p  \n");
		scr_printf("[X] + [<]/[>]   HDTV    720p/1080i [Sq]+ [<]/[>] VGA 448p60 / 480p72  \n");
		scr_printf("[Sel]+[<]/[>] VGA 480p75 / 768p75  [Sel]+[^]     SYNCV VGA 480p60\n");
		scr_printf("----------Preset Commands----------\n");
		scr_printf("    DPad-[v]    => Activate Presets\n");   
		scr_printf("JOY LX,LY,RX,RY => SET DX,DY,DW,DH \n");
		scr_printf("+[L1]/none/[R1] => joy_step 1/4/16 \n");
		scr_printf("[L2] + [<]/[>]  => Mode -16 / +16  \n");
		scr_printf("[R2] + [<]/[>]  => Mode  -1 / +1   \n");
		scr_printf("  [L1] + [R1]   => Adaption: %s    \n",on_off_s[adapt2]);
		scr_printf("  [L2] + [R2]   => SMODE2 = 0x%02lx \n", smode2);
		scr_printf("----------Exit Commands----------\n");
		scr_printf("[START] + [<] / [>]  Launches:   \n");
		scr_printf("  PS2 BROWSER / HDLoader         \n");

		delay(5);
		
		updateflag = -1;
		
		while (updateflag == -1) {

			while(!(waitAnyPadReady(), readpad(), new_pad)); //await a new button
			retval = paddata;
			
			if(paddata & PAD_JOY){
				if(paddata & PAD_L1)
					joy_step = 1;
				else if(paddata & PAD_R1)
					joy_step = 16;
				else
					joy_step = 4;
			}else
				joy_step = 0;

			/* Usual pre-def settings */
			if((retval & PAD_LEFT)&&(retval & PAD_CIRCLE)) {
				interlace = 0x01;			// Interlaced
				mode = 0x02;					// NTSC 480i
				field = 0x00;					// Field mode (Read Every Other Line)
				gs_dx=700; gs_dy=46; gs_magh=3; gs_magv=0;
				gs_dw = 2560; gs_dh = 448;
				syncv = 0;
				smode2 = 1;
				updateflag = 1;
			}
			if((retval & PAD_RIGHT)&&(retval & PAD_CIRCLE)) {
				interlace = 0x00;			// Non-interlaced
				mode = 0x50;					// HDTV-480p
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=272; gs_dy=48; gs_magh=1; gs_magv=0;
				gs_dw = 1360; gs_dh = 480;
				syncv = 0x00C78C0001E00006;
				smode2 = 2;
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_TRIANGLE)) {
				interlace = 0x01;			// Interlaced
				mode = 0x03;					// PAL 576i
				field = 0x00;					// Field mode (Read Every Other Line)
				gs_dx=720; gs_dy=68; gs_magh=3; gs_magv=0;
				gs_dw = 2560; gs_dh = 512;
				syncv = 0;
				smode2 = 1;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_TRIANGLE)) {
				interlace = 0x00;			// Non-interlaced
				mode = 0x53;					// HDTV-576p
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=320; gs_dy=64; gs_magh=1; gs_magv=0;
				gs_dw = 1312; gs_dh = 576;
				syncv = 0;
				smode2 = 2;
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_CROSS))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x52;					// HDTV-720P
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=336; gs_dy=44; gs_magh=0; gs_magv=0;
				gs_dw = 1280; gs_dh = 720;
				syncv = 0x00AB400001400005;
				smode2 = 2;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_CROSS))	{
				interlace = 0x01;			// Interlaced
				mode = 0x51;					// HDTV-1080i
				field = 0x00;					// Field mode (Read Every Other Line)
				gs_dx=248; gs_dy=48; gs_magh=2; gs_magv=1;
				gs_dw = 1920; gs_dh = 1080;
				syncv = 0x0150E00201C00005;
				smode2 = 1;
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_SQUARE))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x1A;					// VGA 448p @ 60Hz
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=350; gs_dy=40; gs_magh=1; gs_magv=0;
				gs_dw = 1024; gs_dh = 448;
				syncv = 0;
				smode2 = 2;
//				syncv = 0x004F80000210000A;
//				smode2 = 1;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_SQUARE))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x1B;					// VGA 480p @ 72Hz
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=360; gs_dy=30; gs_magh=1; gs_magv=0;
				gs_dw = 1152; gs_dh = 480;
				syncv = 0;
				smode2 = 2;
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_SELECT))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x1C;					// VGA 480p @ 75Hz
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=360; gs_dy=18; gs_magh=1; gs_magv=0;
				gs_dw = 1152; gs_dh = 480;
				syncv = 0;
				smode2 = 2;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_SELECT))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x3D;					// VGA 768p @ 75Hz
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=256; gs_dy=30; gs_magh=0; gs_magv=0;
				gs_dw = 1024; gs_dh = 768;
				syncv = 0x006C000001C00001;
				smode2 = 2;
				updateflag = 1;
			}
			else if((retval & PAD_UP)&&(retval & PAD_SELECT))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x1A;					// VGA 640@ 60Hz
				field = 0x01;					// Frame mode (Read Every Line)
				//0037F4FF08828113
				gs_dx=275; gs_dy=40; gs_magh=1; gs_magv=1;
				gs_dw = 1280; gs_dh = 896;
				syncv = 0x004F80000210000A;
				smode2 = 1;
				updateflag = 1;
			}
			else if(retval == PAD_DOWN)	{
				updateflag = 1; //exit inner loop and activate presets
			}
			else if(paddata & (PAD_L3_H0 | PAD_L3_H1))	{
				gs_dx += (1-2*(!(paddata & PAD_L3_H1)))*joy_step;
				gs_dx &= (1<<12)-1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if(paddata & (PAD_L3_V0 | PAD_L3_V1))	{
				gs_dy += (1-2*(!(paddata & PAD_L3_V1)))*joy_step;
				gs_dy &= (1<<11)-1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if(paddata & (PAD_R3_H0 | PAD_R3_H1))	{
				gs_dw += (1-2*(!(paddata & PAD_R3_H1)))*joy_step;
				gs_dw &= (1<<12)-1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if(paddata & (PAD_R3_V0 | PAD_R3_V1))	{
				gs_dh += (1-2*(!(paddata & PAD_R3_V1)))*joy_step;
				gs_dh &= (1<<11)-1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_L2))	{
				mode -= 16;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_L2))	{
				mode += 16;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_R2))	{
				mode -= 1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_R2))	{
				mode += 1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_L1)&&(retval & PAD_R1))	{
				set_kmode();
				Adapt_Flag_2 ^= 1;			//Toggle byte flag
  			set_umode();
				updateflag = 1;
			}
			else if((retval & PAD_L2)&&(retval & PAD_R2))	{
				set_kmode();
				smode2 ^= 3;			//Toggle smode2 interlaced/Progressive
  			set_umode();
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_START))	{
				updateflag = 0;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_START))	{
				updateflag = 0;
			}
		}
	}

	if((retval & PAD_LEFT)&&(retval & PAD_START))	{
		CleanUp();
		scr_printf("\nExiting to PS2 Browser...\n");
		delay(5);
		__asm__ __volatile__(
		".set noreorder\n\t"
		"li $3, 0x04\n\t"
		"syscall\n\t"
		"nop\n\t"
		".set reorder\n\t"
		);
	}
	else if((retval & PAD_RIGHT)&&(retval & PAD_START))	{
		CleanUp();
		strcpy(fullpath, PS2ELF2);
		scr_printf("\nLoading %s...\n", fullpath);
		delay(5);
	   	RunLoaderElf(fullpath, party);	// Run ELF
	}
   
   	SleepThread();					// Should never get here
   
   	return 0;
}

// End of program, leave a blank line afterwards to let 'make' line-command EE-GCC (Emotion Engine - Gnu Compiler Collection) happy ;-)

